module App exposing (..)

import Browser
import Dict exposing (Dict)
import Html as H exposing (Html)
import Html.Attributes as HA
import Html.Events as HE
import Json.Decode as JD
import List.Extra as LE
import Parser exposing ((|=), (|.), Parser)
import Random as R
import Random.Extra as RE
import Set
import Svg
import Svg.Attributes as SA
import Svg.Events as SE
import Tuple exposing (first, second, pair, mapSecond)
import Util exposing (..)
import Vector exposing (..)


recipes : List Recipe
recipes = List.filterMap (Parser.run parse_recipe >> Result.toMaybe)
  [ "⛺1 🧍1🌲1🌿1 🛖 🧍1"
  , ". 🧍1🌲1 ⛺1 🧍1"
  , ". 🧍1🌲1 . 🧍1🔥1"
  , "🛖1 🧍1🪨1🔥1 🛖 🧍1🧱"
  , "⛺1 🧍1🧱3🌲2  🏠1 🧍1"
  , "🏠3 . 🏘️1 ."
  , "🏠1 🍞3 . 🧍1"
  , ". 🧍1 . 🧍1🪨1"
  , ". 🧍1🌲1 . 🧍1🌱2"
  , ". 🌱1 . 🌲1"
  , ". 🔥1🌱1 . 🌿1"
  , ". 🔥1🌿1 . 🍞1"
  ]

item_kinds =
  [ "🌲"
  , "🌱"
  , "🌿"
  , "🪨"
  , "🧱"
  , "🔥"
  , "🧍"
  , "🍞"
  ]

building_kinds =
  [ "⛺"
  , "🛖"
  , "🏠"
  ]

main = Browser.document
    { init = init
    , update = update
    , subscriptions = subscriptions
    , view = view
    }

type alias TownId = Int

type alias Path = List Vector

type ClickMode
  = BlankClick
  | DragRoad TownId (Path)
  | DragItem Item (List TownId) (Maybe (Road, Float))
  | PanCamera Vector

type Tool
  = RoadTool
  | MoveItemTool
  | MineTool

type alias Model = 
  { pointer : Vector
  , click_mode : ClickMode
  , tool : Tool
  , map : Map
  , selected_town : Maybe TownId
  , debug_log : List String
  , camera : Camera
  }

type alias Camera =
  { pan : Vector
  , zoom : Float
  }

init_camera = 
  { pan = (0,0)
  , zoom = 1
  }

type Road = Road TownId TownId (Path)

type alias Obstacle =
  { position : Vector
  , radius : Float
  }

type alias ItemKind = String

type alias Item = 
  { town : TownId
  , kind : ItemKind
  }

type alias BuildingKind = String

type alias Recipe =
  { in_buildings : List (BuildingKind,Int)
  , in_items : List (ItemKind,Int)
  , out_buildings : List (BuildingKind,Int)
  , out_items : List (ItemKind,Int)
  }

type alias Town =
  { position : Vector
  , id : TownId
  , items : List Item
  , buildings : List BuildingKind
  , name : String
  }

type alias Map =
  { towns : List Town
  , obstacles : List Obstacle
  , roads : List Road
  }

parse_symbol_list : Parser (List (String,Int))
parse_symbol_list =
  Parser.loop 
    []
    (\l -> 
      Parser.oneOf
        [ Parser.succeed 
            (\(symbol,count) -> Parser.Loop (if symbol=="." then l else (symbol,count)::l))
            |= (   Parser.succeed pair
                |= Parser.variable 
                    { start = \c -> not (Char.isDigit c || c == ' ')
                    , inner = \c -> not (Char.isDigit c || c == ' ')
                    , reserved = Set.empty
                    }
                 |= Parser.oneOf [ Parser.int, Parser.succeed 1 ]
               )
        , Parser.succeed () |> Parser.map (\_ -> Parser.Done (List.reverse l))
        ]
    )

parse_recipe : Parser Recipe
parse_recipe 
  = Parser.succeed Recipe
    |= parse_symbol_list
    |. Parser.keyword " "
    |. Parser.spaces
    |= parse_symbol_list
    |. Parser.keyword " "
    |. Parser.spaces
    |= parse_symbol_list
    |. Parser.keyword " "
    |. Parser.spaces
    |= parse_symbol_list
  

blank_map = 
  { towns = []
  , obstacles = []
  , roads = []
  }

init_model = 
  { pointer = (0,0)
  , click_mode = BlankClick
  , tool = RoadTool
  , map = blank_map
  , selected_town = Nothing
  , debug_log = []
  , camera = init_camera
  }

type Msg
  = SetPosition Vector
  | PointerDown
  | PointerUp
  | TouchStart Vector
  | TouchEnd
  | SetTool Tool
  | SetMap Map
  | ApplyRecipe TownId Recipe
  | EndTurn
  | BadValue JD.Value

init : () -> (Model, Cmd Msg)
init _ = (init_model, R.generate SetMap map_generator)

debug : String -> Model -> Model
debug msg model = { model | debug_log = msg::model.debug_log }

nocmd model = (model, Cmd.none)

decode_vector : JD.Decoder Vector
decode_vector = JD.map2 pair (JD.field "x" JD.float) (JD.field "y" JD.float)

decode_event_vector : (Vector -> a) -> JD.Decoder a
decode_event_vector fn = 
  JD.map fn (JD.field "detail" decode_vector)

decode_touch_vector : (Vector -> a) -> JD.Decoder a
decode_touch_vector fn =
  JD.map fn (JD.at ["detail", "0", "position"] decode_vector)

poisson : Float -> R.Generator Int
poisson lambda =
  let
    l = e ^ (-lambda)

    step k p =
         R.float 0 1
      |> R.andThen 
          (\u ->
            let
              np = p * u
            in
              if np <= l then R.constant (k-1) else step (k+1) np
          )
  in
    step 0 1

projected_pointer : Model -> Vector
projected_pointer model = sub model.pointer model.camera.pan |> smul model.camera.zoom

get_town : Model -> TownId -> Maybe Town
get_town model i = LE.getAt i model.map.towns

next_turn : Map -> R.Generator Map
next_turn original_map =
  let
    turn_town : Town -> R.Generator Town
    turn_town town =
      let
        num_items = List.length town.items
        max_items = 3
        rate = 0.9

        num_new_items : R.Generator Int
        num_new_items = poisson rate |> R.map (min (max_items - num_items))

        random_item_kinds = 
          [ "🌲"
          , "🌿"
          ]
        
        (k1, rest_kinds) = case random_item_kinds of
          [] -> ("",[])
          a::rest -> (a,rest)

        new_item : R.Generator Item
        new_item = 
          R.map2 Item
            (R.constant town.id)
            (R.uniform k1 rest_kinds)

        new_items : R.Generator (List Item)
        new_items = num_new_items |> R.andThen (\n -> R.list n new_item)

        add_new_items t = new_items |> R.map (\items -> { t | items = items++t.items })

        available_recipes : List Recipe
        available_recipes = 
             recipes
          |> List.filter (can_afford_recipe town)
      in
           R.constant town
        |> R.andThen add_new_items

    turn_towns map =
         RE.sequence (List.map turn_town map.towns)
      |> R.map (\towns -> { map | towns = towns })
  in
       R.constant original_map
    |> R.andThen turn_towns

easy_uniform : a -> List a -> R.Generator a
easy_uniform never options = case options of
  [] -> R.constant never
  a::rest -> R.uniform a rest

random_name : R.Generator String
random_name =
  let
    vowels = ["a","e","i","o","u"]
    consonants = ["b","c","d","g","m","n","p","r","s","t","ch","th","sh"]

    uniform_string = easy_uniform ""

    random_syllable = R.map2 (++) (uniform_string consonants) (uniform_string vowels)
  in
       R.list 3 random_syllable
    |> R.map (String.join "")

map_generator : R.Generator Map
map_generator =
  let
    random_position : R.Generator Vector
    random_position = R.map2 pair (R.float -20 20) (R.float -20 20)

    random_town : R.Generator Town
    random_town = 
      R.map2 (\p name -> { position = p, id = 0, items = [], buildings = ["⛺"], name = name })
        random_position
        random_name

    random_obstacle =
      R.map2 (\p size -> { position = p, radius = size })
        random_position
        (R.float (2^2) (6^2) |> R.map sqrt)

    generate_obstacles map = 
         R.list 20 random_obstacle 
      |> R.map (\o -> { map | obstacles = o })

    generate_towns : Map -> R.Generator Map
    generate_towns map = 
         R.list 10 random_town 
      |> R.map (List.indexedMap (\i t -> { t | id = i })) 
      |> R.map (distribute_towns map.obstacles)
      |> R.map (\t -> { map | towns = t })

    add_to_random_town : ItemKind -> Map -> R.Generator Map
    add_to_random_town kind map = 
      case map.towns of
        [] -> R.constant map
        t1::tothers ->
          let
            rtown = R.uniform t1 tothers
          in
            rtown
            |> R.map (\town -> { map | towns = LE.setIf (.id >> (==) town.id) { town | items = ({ kind = kind, town = town.id }::town.items) } map.towns })

  in
       R.constant blank_map
    |> R.andThen generate_obstacles
    |> R.andThen generate_towns
    |> R.andThen (add_to_random_town "🧍")
    |> R.andThen (add_to_random_town "🌲")

add_item : Model -> R.Generator (Maybe TownId)
add_item model = case model.map.towns of 
  [] -> R.constant Nothing
  a::rest -> R.uniform a rest |> R.map (.id >> Just)

fract : Float -> (Int,Float)
fract x = (floor x, x - (tf (floor x)))

total_length points = List.map2 (\v1 v2 -> distance v1 v2) points (List.drop 1 points) |> List.sum

simplify_path : Float -> Path -> Path
simplify_path step points =
  let
    simplify_middle : Vector -> (Vector,Float,Path) -> (Vector,Float,Path)
    simplify_middle p (last,r,opoints) =
      let
        v = sub p last
        d = len v
        n = normalise v
        (steps,nr) = fract ((r + d) / step)
        nopoints = List.map (\k -> add last (smul (((tf k)-r)*step) n)) (List.range 1 steps)
      in
        (p,nr,opoints ++ nopoints)
  in
    case (points, List.reverse points) of
      (start::rest, end::_) -> start::(List.foldl simplify_middle (start,0,[]) points |> \(_,_,opoints) -> opoints)++[end]
      _ -> []

simplify_road = simplify_path 1

set_pointer pos model = { model | pointer = pos }

update_click_mode : Model -> Model
update_click_mode model = 
  let
    pointer = projected_pointer model
  
    mode = case model.click_mode of
      BlankClick -> BlankClick

      DragRoad i path -> DragRoad i (pointer::path)

      PanCamera v -> PanCamera model.pointer

      DragItem item route mroad -> 
        case mroad of
            Nothing -> 
              let
                mcp = drag_item_position model item
                mt = get_town model item.town
                nmroad = 
                  case mcp of
                    Nothing -> Nothing
                    Just (Road _ _ path as r,v) -> 
                      let
                        md = Maybe.map (.position >> distance v) mt
                        dr = distance_of_point_on_road r v
                        total = total_length path
                      in
                           Maybe.map2 pair mt md 
                        |> Maybe.andThen 
                            (\(t,d) -> 
                              if dr > 1 && dr < total - 1 then 
                                (Just (r, dr)) 
                              else
                                Nothing
                            )
                nroute = case nmroad of
                  Nothing -> route
                  Just (Road from to _, _) -> case route of
                    a::b::rest -> 
                      if a==item.town && ((a==from && b==to) || (a==to &&b==from)) then
                        b::rest
                      else
                        route
                    _ -> route
                    
              in
                DragItem item nroute nmroad
            Just (Road from to path as r,d) -> 
              let
                mt1 = get_town model from
                mt2 = get_town model to
                mcp = closest_point_to_road r pointer
                mclosest = 
                  [from,to]
                  |> List.filterMap (get_town model >> Maybe.map2 (\cp t -> (t,distance cp t.position)) mcp)
                  |> List.sortBy second
                  |> List.head
                nd = Maybe.map (distance_of_point_on_road r) mcp |> Maybe.withDefault d
                max_move = 3
                nnd = d + (clamp (-max_move) max_move (nd-d))
              in
                case mclosest of
                  Nothing -> DragItem item route (Just (r,nnd))
                  Just (t,td) -> if td < 1 then DragItem {item | town = t.id } (if List.head route == Just t.id then route else t.id::route) Nothing else DragItem item route (Just (r,nnd))
  in
    case model.click_mode of
      PanCamera p1 ->
        let
          d = sub model.pointer p1 |> smul model.camera.zoom

          ncamera = 
            { pan = add d model.camera.pan
            , zoom = model.camera.zoom
            }
        in
          { model | click_mode = mode, camera = ncamera }
          
      _ -> { model | click_mode = mode }

update_map : (Map -> Map) -> Model -> Model
update_map fn model = { model | map = fn model.map }

update_town : (Town -> Town) -> TownId -> Model -> Model
update_town fn townid = update_map (\map -> { map | towns = LE.updateIf (.id >> (==) townid) fn map.towns })

change_selected_town : Model -> Model
change_selected_town model = 
  { model | selected_town = select_town model |> Maybe.map (third >> .id) }

stock_of_item : String -> Town -> Int
stock_of_item kind = .items >> List.filter (.kind >> (==) kind) >> List.length

spend_buildings : String -> Int -> TownId -> Model -> Model
spend_buildings kind num = 
  update_town (\town -> { town | buildings = List.foldl (\building (n,o) -> if building == kind && n>0 then (n-1,o) else (n,building::o)) (num,[]) town.buildings |> second })

spend_items : String -> Int -> TownId -> Model -> Model
spend_items kind num = 
  let
    q = Debug.log "spend" (kind,num)
  in
    update_town (\town -> { town | items = List.foldl (\item (n,o) -> if item.kind == kind && n>0 then (n-1,o) else (n,item::o)) (num,[]) town.items |> second })

end_touch : Model -> Model
end_touch model = case (Debug.log "end_touch" model.click_mode) of
  BlankClick -> model |> change_selected_town

  DragRoad id1 path -> 
    let
    
      road = case select_town model of
        Nothing -> Err "no town"
    
        Just (j,d,t) -> 
          if id1==t.id then 
            Err "same start and end"
          else
            Ok <| Road id1 t.id (simplify_road (t.position::path))

      is_valid rroad = case rroad of
        Road _ _ rpath -> if valid_road model rpath then Ok rroad else Err "hits an obstacle"

      cost = road_cost path
            
      can_afford rroad = case rroad of
        Road from to _ -> 
          let
            rocks_from = from |> get_town model |> Maybe.map (stock_of_item "🪨") |> Maybe.withDefault 0
          in
            if rocks_from >= cost then
              Ok rroad
            else
              Err "Not enough 🪨"

      result = 
           road
        |> Result.andThen is_valid
        |> Result.andThen can_afford

      finish_road m =
        case result of
          Err _ -> m
          Ok rroad -> 
               m 
            |> update_map (\map -> { map | roads = rroad::map.roads })
            |> spend_items "🪨" cost id1
    in
         model
      |> finish_road
      |> change_selected_town
      

  DragItem item route _ -> 
    let
      q = Debug.log "drop" item
      ntowns = LE.updateAt item.town (\t -> { t | items = item::t.items }) model.map.towns
    in
      { model | selected_town = Just item.town } |> debug (Debug.toString item) |> update_map (\map -> { map | towns = ntowns })

  PanCamera _ -> model

set_blank_click : Model -> Model
set_blank_click model = { model | click_mode = BlankClick }

select_town : Model -> Maybe (Int, Float, Town)
select_town model = closest_within (\t -> 2) .position (projected_pointer model) model.map.towns

select_item : Model -> Maybe (Item, Int, Vector)
select_item model =
  let
    items = positioned_items model.map
  in
       closest_within (\_ -> 1.5) third (projected_pointer model) items
    |> Maybe.map third
    
roads_at_town : Model -> TownId -> List Road
roads_at_town model id = 
     model.map.roads
  |> List.filter 
      (\r -> case r of
        Road from to _ -> from==id || to==id
      )

drag_item_position : Model -> Item -> Maybe (Road,Vector)
drag_item_position model item =
  let
    roads = roads_at_town model item.town 
    pointer = projected_pointer model
  in
         List.filterMap (\r -> closest_point_to_road r pointer |> Maybe.map (pair r)) roads
      |> List.sortBy (second >> sub pointer >> len)
      |> List.head

oneOfMaybes : List (Maybe a) -> Maybe a
oneOfMaybes maybes = case maybes of
  [] -> Nothing
  Nothing::rest -> oneOfMaybes rest
  (Just a)::rest -> Just a

positioned_items : Map -> List (Item, Int, Vector)
positioned_items map =
     map.towns
  |> List.concatMap 
     (\t -> t.items |> List.indexedMap (\i item -> (item, i, item_position t item.kind |> add t.position)))
    
update : Msg -> Model -> (Model, Cmd Msg)
update msg model = case msg of
  SetPosition pos -> model |> set_pointer pos |> update_click_mode |> nocmd

  PointerDown -> 
    let
      towns = model.map.towns

      (mode, ntowns) = oneOfMaybes
        [ select_item model |> Maybe.map (\(item,i,pos) -> (DragItem item [item.town] Nothing, LE.updateAt item.town (\t -> { t | items = LE.removeAt i t.items }) model.map.towns))
        , select_town model |> Maybe.map (\(i,d,town) -> (DragRoad town.id [town.position], towns))
        ]
        |> Maybe.withDefault (PanCamera model.pointer, towns)
      
    in
      case model.click_mode of
        BlankClick -> 
          { model | click_mode = mode} |> debug "PointerDown" |> update_map (\map -> { map | towns = ntowns }) |> nocmd

        _ -> model |> nocmd

  PointerUp -> model |> debug "PointerUp" |> end_touch |> set_blank_click |> nocmd

  TouchStart pos -> 
    let
      q = Debug.log "TouchStart" pos
    in
      { model | pointer = pos } |> debug "TouchStart" |> update PointerDown

  TouchEnd -> 
    let
      q = Debug.log "TouchEnd" ()
    in
      update PointerUp (model |> debug "TouchEnd")

  SetTool tool -> { model | tool = tool } |> nocmd

  SetMap map -> 
    let
      person_town = 
           map.towns 
        |> LE.find (.items >> List.any (.kind >> (==) "🧍"))
        
      camera = model.camera |> \c -> case person_town of
        Just t -> { c | pan = sub (0,0) t.position }
        Nothing -> c
    in
      { model | map = map, camera = camera } |> nocmd

  ApplyRecipe townid recipe -> model |> apply_recipe townid recipe |> nocmd

  EndTurn -> (model, R.generate SetMap (next_turn model.map))

  BadValue v -> case JD.decodeValue (decode_touch_vector TouchStart) v of
    Ok _ -> model |> nocmd
    Err err -> 
      let
        q = Debug.log "err" err
      in
        model |> nocmd

subscriptions model = Sub.none

apply_recipe : TownId -> Recipe -> Model -> Model
apply_recipe townid recipe model = case get_town model townid of
  Nothing -> model
  Just town ->
    let
      can_afford = can_afford_recipe town recipe

      spent_buildings : Model -> Model
      spent_buildings m =
        List.foldl
          (\(b,c) buildings -> spend_buildings b c townid buildings)
          m
          recipe.in_buildings

      spent_items : Model -> Model
      spent_items m =
        List.foldl
          (\(b,c) items -> spend_items b c townid items)
          m
          recipe.in_items
          
      new_buildings : List BuildingKind
      new_buildings = List.concatMap (\(b,c) -> List.repeat c b) recipe.out_buildings

      new_items : List Item
      new_items = List.concatMap (\(b,c) -> List.repeat c { town = townid, kind = b }) recipe.out_items

      spent_model = 
           model 
        |> spent_buildings
        |> spent_items
        |> update_town (\t -> { t | buildings = new_buildings++t.buildings, items = new_items++t.items}) townid

      max_items = 500

      final_items =
           get_town spent_model townid
        |> Maybe.map (.items >> List.length)
        |> Maybe.withDefault 0
    in
      if can_afford && final_items <= max_items then
        spent_model
      else
        model

iterate : (a -> a) -> Int -> a -> a
iterate fn times list = List.foldl (\_ -> fn) list (List.range 1 times)

iterate_until : Int -> (a -> a) -> (a -> a -> Bool) -> a -> a
iterate_until run fn condition list =
  let
    nlist = fn list
  in
    if condition list nlist then nlist else (iterate_until (run + 1) fn condition nlist)

distribute_dots : Float -> List Obstacle -> Path
distribute_dots min_dist dots =
  let
    force_for_dot dot =
      let
        p1 = dot.position
        r1 = dot.radius
        dot_forces = 
             dots 
          |> List.filter ((/=) dot)
          |> List.map 
            (\d2 ->
              let
                p2 = d2.position
                r2 = d2.radius
                v = sub p2 p1
                d = (len v) - (min_dist + r1 + r2)
                n = normalise v
                force_strength = if d < 0 then 0.1 else 0
              in
                smul (d*force_strength) n
            )
      in
        sum dot_forces
        
    forces = List.map force_for_dot dots
  in
    List.map2 (\d f -> add d.position f) dots forces

distribute_towns : List Obstacle -> List Town -> List Town
distribute_towns obstacles original_towns = 
  let
    step towns = 
      let
        town_obstacles : List Obstacle
        town_obstacles = List.map (\t -> {position = t.position, radius = 1}) towns
      
        positions : Path
        positions = distribute_dots 5 (town_obstacles ++ obstacles)
      in
        List.map2 (\t p -> { t | position = p }) towns positions

    stopped_moving otowns ntowns =
         List.map2 (\t1 t2 -> distance t1.position t2.position) otowns ntowns
      |> List.sum
      |> (>) 1
  in
    iterate_until 0 step stopped_moving original_towns

closest_point_to_road : Road -> Vector -> Maybe Vector
closest_point_to_road road p = case road of
  Road _ _ path ->
       List.map2 pair path (List.drop 1 path) 
    |> List.map (\segment -> closest_point_on_segment segment p |> second)
    |> LE.minimumBy (\cp -> len (sub p cp))

point_at_distance_on_road : Road -> Float -> Maybe Vector
point_at_distance_on_road road d = case road of
  Road _ _ path -> point_at_distance_on_path path d

point_at_distance_on_path : Path -> Float -> Maybe Vector
point_at_distance_on_path path d = case path of
  [] -> Nothing
  a::[] -> Just a
  a::b::rest ->
    let
      v = sub b a
      l = distance a b
    in
      if d < l then Just (add a (smul (d/l) v)) else point_at_distance_on_path (b::rest) (d-l)

distance_of_point_on_road : Road -> Vector -> Float
distance_of_point_on_road road p = case road of
  Road _ _ original_path -> 
    let
      fn d path = case path of
        [] -> 0
        a::[] -> d
        a::b::rest ->
          let
            (t,cp) = closest_point_on_segment (a,b) p
            dp = distance p cp
            l = distance a b
          in
            if dp < 0.1 then d + t*l else fn (d+l) (b::rest)
    in
      fn 0 original_path

d_for_path : Path -> String
d_for_path = 
     List.indexedMap (\i (x,y) -> (if i==0 then "M" else "L")++" "++(ff x)++" "++(ff y))
  >> String.join " "

translate_to : Vector -> String
translate_to position = 
     "translate("
  ++ (ff <| first position)
  ++ " "
  ++ (ff <| second position)
  ++" )"

path_intersects_obstacle path obstacle = 
  List.any (distance obstacle.position >> (>) obstacle.radius) path
    
valid_road : Model -> Path -> Bool
valid_road model path =
  let
    map = model.map

    hit_obstacles = 
      map.obstacles
      |> List.filter (path_intersects_obstacle path)
  in
    hit_obstacles == []

road_cost : Path -> Int
road_cost = total_length >> (*) 0.2 >> ceiling

item_position : Town -> String -> Vector
item_position town kind =
  let
    i = LE.elemIndex kind item_kinds |> Maybe.withDefault 0
    an = (tf i) / (tf <| List.length item_kinds - 1) * (2*pi/3) + (pi/6)
    r = 2
    x = r * (cos an)
    y = r * (sin an)
  in
    (x,y)

item_icon kind = icon 2 kind []

icon : Float -> String -> (List (H.Attribute msg)) -> Svg.Svg msg
icon size text attrs =
  Svg.text_
    (  attrs
    ++ [ SA.fontSize <| ff size
       , SA.dominantBaseline "middle"
       , SA.textAnchor "middle"
       ]
    )
    [ Svg.text text ]

count_kinds : List comparable -> List (comparable,Int)
count_kinds =
     List.sort
  >> LE.group
  >> List.map (Tuple.mapSecond (List.length >> (+) 1))

stock_counts : Town -> List (ItemKind, Int)
stock_counts town = 
     town.items
  |> List.map .kind
  |> count_kinds

building_counts : Town -> List (BuildingKind, Int)
building_counts town =
     town.buildings
  |> count_kinds

has_enough : a -> Int -> List a -> Bool
has_enough kind count list =
     list
  |> List.filter ((==) kind)
  |> List.length
  |> (<=) count

has_enough_items : ItemKind -> Int -> Town -> Bool
has_enough_items kind count =
     .items
  >> List.map .kind
  >> has_enough kind count

has_enough_buildings : BuildingKind -> Int -> Town -> Bool
has_enough_buildings kind count =
     .buildings
  >> has_enough kind count

can_afford_recipe : Town -> Recipe -> Bool
can_afford_recipe town recipe =
  let
    enough_buildings = List.all (\(b,c) -> has_enough_buildings b c town) recipe.in_buildings
    enough_items = List.all (\(b,c) -> has_enough_items b c town) recipe.in_items

  in
    enough_items && enough_buildings

phi = (sqrt 5 + 1) / 2

inverse_triangle : Int -> Int
inverse_triangle k = 
  (sqrt ((tf k) * 8 + 1))/2 - 0.5 |> floor

spiral : Float -> Int -> Vector
spiral scale i =
  let
    an = (tf i) * phi/4 - phi/2
    r = scale * (tf i)
  in
    (r * (cos an), r * (sin an))

fan : Int -> Vector
fan n =
  let
    r = inverse_triangle n
    t = r*(r+1) // 2
    m = n - t
    dx = 2
    dy = 0.5
  in
    (-dx*(tf r)/2 + dx*(tf m),-dy*(tf r))
  

view model = 
  let
    view_town town =
      let
        num_items = List.length town.items
        item_stocks = stock_counts town
        view_item (kind, count) =
          let
            pos = item_position town kind
          in
            Svg.g
              [ SA.transform <| translate_to pos ]
              [ item_icon kind
              , icon 1 (fi count) 
                  [ SA.fill "white"
                  , SA.stroke "black"
                  , SA.strokeWidth "0.2"
                  ]
              ]
        items = 
          item_stocks
          |> List.map view_item
          |> Svg.g []

        buildings =
             town.buildings
          |> List.sortBy (\b -> LE.elemIndex b building_kinds |> Maybe.withDefault 0)
          |> List.reverse
          |> List.indexedMap 
              (\i b -> 
                let
                  pos = fan i
                in
                  icon 3 b [ SA.transform <| translate_to pos ]
              )
          |> List.reverse
          |> Svg.g []

        name =
          Svg.text_
            [ SA.textAnchor "middle" 
            , SA.fontSize "1"
            , SA.y "-1.5"
            ]
            [ Svg.text <| capitalise town.name ]
      in
        Svg.g
          [ SA.transform <| translate_to town.position
          ]
          [ buildings
          , name
          , items
          ]

    view_road road = case road of
      Road from to path ->
        Svg.g
          [ SA.class "road" ]
          [ Svg.path
            [ SA.d <| d_for_path <| path
            , SA.fill "none"
            , SA.stroke "oklch(0.9 0 0)"
            ]
            []
          , Svg.path
            [ SA.d <| d_for_path <| path
            , SA.fill "none"
            , SA.stroke "oklch(1 0 0)"
            , SA.strokeWidth "0.2"
            , SA.strokeDasharray "1 1"
            ]
            []
          ]

    view_obstacle obstacle =
      Svg.g
        [ SA.transform <| translate_to obstacle.position ]
        [ Svg.circle
            [ SA.r <| ff obstacle.radius 
            , SA.fill "blue"
            ]
            []
        ]

    view_click_mode = case model.click_mode of
      BlankClick -> Svg.text ""

      DragRoad townid path ->
        let
          valid = valid_road model path
        in
          view_road (Road 0 0 (simplify_road (List.reverse path)))

      DragItem item route mroad -> 
        let
          mcp = case mroad of
            Nothing -> drag_item_position model item |> Maybe.map second
            Just ((Road from to path) as r, d) -> point_at_distance_on_road r d
        in
          case mcp of
            Nothing -> Svg.text ""
            Just cp -> 
              Svg.g
                [ SA.transform <| translate_to cp ]
                [ Svg.circle
                  [ SA.r "0.5"
                  , SA.fill "lightgrey"
                  ]
                  []
                , item_icon item.kind
                ]

      PanCamera _ -> Svg.text ""

    p1 = (12,3)
    p2 = (-15,0)

    tool_selector (tool,label) =
      H.label
        []
        [ H.text label
        , H.input
          [ HE.onClick (SetTool tool)
          , HA.type_ "radio"
          , HA.checked <| model.tool == tool
          ]
          [ H.text label ]
        ]

    selected_town =
         model.selected_town
      |> Maybe.andThen (\i -> LE.getAt i model.map.towns)

    selected_town_contents = case selected_town of
      Nothing -> [H.text ""]
      Just town ->
        let
          building_stock = Dict.fromList (building_counts town)
          item_stock = Dict.fromList (stock_counts town)

          view_recipe recipe =
            let
              first_two (a,b,c) = (a,b)
            
              show_count : (String,Int) -> String
              show_count (b,n) = List.repeat n b |> String.join ""
            
              in_building_stocks : List (BuildingKind, Int, Int)
              in_building_stocks =
                   recipe.in_buildings
                |> List.map (\(b,n) -> (b, n, Dict.get b building_stock |> Maybe.withDefault 0))
            
              in_buildings = 
                   in_building_stocks
                |> List.map (first_two >> show_count)
                   
              in_item_stocks = 
                   recipe.in_items
                |> List.map (\(b,n) -> (b, n, Dict.get b item_stock |> Maybe.withDefault 0))

              in_items =
                   in_item_stocks
                |> List.map (first_two >> show_count)

              out_buildings =
                   recipe.out_buildings
                |> List.map (show_count)

              out_items =
                   recipe.out_items
                |> List.map (show_count)
            in
              H.li
                [ HE.onClick (ApplyRecipe town.id recipe)]
                [ H.text <| String.join " " in_buildings
                , H.text <| String.join " " in_items
                , H.text " → "
                , H.text <| String.join " " out_buildings
                , H.text <| String.join " " out_items
                ]
        
          available_recipes = 
               recipes
            |> List.filter (can_afford_recipe town)
            |> List.map view_recipe
        in
          [ H.p [] [ H.text <| capitalise <| town.name ]
          , H.ul
              [ HA.id "recipe-list" ]
              available_recipes
          ]
          
    status_bar = 
      H.fieldset
        [ HA.id "status-bar" ]
        (case model.click_mode of
            DragRoad townid path ->
              let
                cost = road_cost path
              in
                [ H.text <| "Building road would cost " ++ (fi cost) ++ "🪨" ]

            DragItem item _ _ -> [ H.text <| "Moving " ++ item.kind ]
      
            BlankClick -> 
              selected_town_contents

            PanCamera v -> [H.text ""]
        )
        
  in
    {
      title = "Hey",
      body = 
        [ Svg.svg
            [ SE.on "svgmove" <| decode_event_vector SetPosition
            , SE.on "mousedown" <| JD.succeed PointerDown
            , SE.on "svgtouchstart" <| JD.oneOf [decode_touch_vector TouchStart, JD.map BadValue JD.value]
            , SE.on "mouseup" <| JD.succeed PointerUp
            , SE.on "svgtouchend" <| JD.succeed TouchEnd
            , HA.attribute "viewBox" "-15 -15 30 30"
            ]
            [ Svg.g
                [ SA.id "map"
                , SA.transform <| translate_to model.camera.pan
                ]
                [ Svg.rect
                    [ SA.x "-1000"
                    , SA.y "-1000"
                    , SA.width "2000"
                    , SA.height "2000"
                    , SA.fill "hsl(120,80%,70%)"
                    ]
                    []
                , Svg.g
                    [ SA.id "obstacle-outlines" ]
                    (List.map view_obstacle model.map.obstacles)
                , Svg.g
                    [ SA.id "obstacles" ]
                    (List.map view_obstacle model.map.obstacles)
                , Svg.g
                    [ SA.id "roads" ]
                    (List.map view_road model.map.roads)
                , Svg.g
                    [ SA.id "towns" ]
                    (List.map view_town model.map.towns)
                , view_click_mode
                ]
            , Svg.rect
                [ SA.fillOpacity "0.0001"
                , SA.x "-10000"
                , SA.y "-10000"
                , SA.width "20000"
                , SA.height "20000"
                ]
                
                []
            ]
        , H.div
          [ HA.id "controls" ]
          [ status_bar
          , H.button
              [ HE.onClick EndTurn ]
              [ H.text "End turn" ]
          ]
        ]
    }